/*
 * Copyright 2017-present Open Networking Foundation
 * Copyright © 2020 camunda services GmbH (info@camunda.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.atomix.cluster.impl;

import static org.assertj.core.api.Assertions.assertThat;

import io.atomix.cluster.BootstrapService;
import io.atomix.cluster.ClusterMembershipEvent;
import io.atomix.cluster.Member;
import io.atomix.cluster.TestBootstrapService;
import io.atomix.cluster.TestGroupMembershipProtocol;
import io.atomix.cluster.discovery.BootstrapDiscoveryProvider;
import io.atomix.cluster.discovery.ManagedNodeDiscoveryService;
import io.atomix.cluster.messaging.impl.TestMessagingServiceFactory;
import io.atomix.cluster.messaging.impl.TestUnicastServiceFactory;
import io.atomix.cluster.protocol.GroupMembershipEvent;
import io.atomix.utils.Version;
import java.util.concurrent.atomic.AtomicReference;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.parallel.Execution;
import org.junit.jupiter.api.parallel.ExecutionMode;

@Execution(ExecutionMode.CONCURRENT)
final class DefaultClusterMembershipServiceTest {

  private final TestMessagingServiceFactory messagingServiceFactory =
      new TestMessagingServiceFactory();
  private final TestUnicastServiceFactory unicastServiceFactory = new TestUnicastServiceFactory();
  private final Member localMember = Member.member("0", "localhost:5000");
  private final Version version = Version.from("1.0.0");
  private final BootstrapService bootstrapService =
      new TestBootstrapService(
          messagingServiceFactory.newMessagingService(localMember.address()),
          unicastServiceFactory.newUnicastService(localMember.address()));
  private final ManagedNodeDiscoveryService discoveryService =
      new DefaultNodeDiscoveryService(
          bootstrapService, localMember, BootstrapDiscoveryProvider.builder().build());

  // keep the type as a test type so we can use its test utilities
  private final TestGroupMembershipProtocol protocol = new TestGroupMembershipProtocol();

  @Test
  void shouldManageDiscoveryService() {
    // given
    final var membershipService =
        new DefaultClusterMembershipService(
            localMember, version, discoveryService, bootstrapService, protocol);

    // when - then
    membershipService.start().join();
    assertThat(discoveryService.isRunning()).as("the discovery service is now running").isTrue();

    // when - then
    membershipService.stop().join();
    assertThat(discoveryService.isRunning())
        .as("the discovery service is not running anymore")
        .isFalse();
  }

  @Test
  void shouldGetLocalMember() {
    // given
    final var membershipService =
        new DefaultClusterMembershipService(
            localMember, version, discoveryService, bootstrapService, protocol);

    // when
    membershipService.start().join();

    // then
    final var expectedMember = new StatefulMember(localMember, version);
    assertThat(membershipService.getLocalMember())
        .as("the local member is member 0")
        .isEqualTo(expectedMember);
  }

  @Test
  void shouldGetMembers() {
    // given
    final var membershipService =
        new DefaultClusterMembershipService(
            localMember, version, discoveryService, bootstrapService, protocol);

    // when
    protocol.getMembers().add(Member.member("1", "localhost:5001"));
    membershipService.start().join();

    // then
    assertThat(membershipService.getMembers())
        .as("there should be at least one member")
        .isNotEmpty()
        .as("the reported members are exactly the same as the protocol's")
        .containsExactlyInAnyOrderElementsOf(protocol.getMembers());
  }

  @Test
  void shouldTrackLocalMemberReachability() {
    // given
    final var membershipService =
        new DefaultClusterMembershipService(
            localMember, version, discoveryService, bootstrapService, protocol);

    // when - then
    membershipService.start().join();
    assertThat(membershipService.getLocalMember().isActive())
        .as("local member is active after starting")
        .isTrue();

    // when - then
    membershipService.stop().join();
    assertThat(membershipService.getLocalMember().isActive())
        .as("local member is not active after stopping")
        .isFalse();
  }

  @Test
  void shouldTrackReachabilityOfLocalMember() {
    // given
    final var membershipService =
        new DefaultClusterMembershipService(
            localMember, version, discoveryService, bootstrapService, protocol);

    // when - then
    membershipService.start().join();
    assertThat(membershipService.getLocalMember().isReachable())
        .as("local member is reachable after starting")
        .isTrue();

    // when - then
    membershipService.stop().join();
    assertThat(membershipService.getLocalMember().isReachable())
        .as("local member is not reachable after stopping")
        .isFalse();
  }

  @Test
  void shouldManageGroupMembershipOfLocalMember() {
    // given
    final var membershipService =
        new DefaultClusterMembershipService(
            localMember, version, discoveryService, bootstrapService, protocol);

    // when - then
    membershipService.start().join();
    assertThat(protocol.getMembers())
        .as("local member has joined the group protocol")
        .containsExactly(membershipService.getLocalMember());

    // when - then
    membershipService.stop().join();
    assertThat(protocol.getMembers()).as("local member has left the group protocol").isEmpty();
  }

  @Test
  void shouldForwardGroupEvents() {
    // given
    final var membershipService =
        new DefaultClusterMembershipService(
            localMember, version, discoveryService, bootstrapService, protocol);
    membershipService.start().join();

    // when
    final var event =
        new GroupMembershipEvent(
            GroupMembershipEvent.Type.MEMBER_REMOVED, Member.member("1", "localhost:5001"));
    final var receivedEvent = new AtomicReference<ClusterMembershipEvent>();
    membershipService.addListener(receivedEvent::set);
    protocol.sendEvent(event);

    // then
    final var expectedEvent =
        new ClusterMembershipEvent(
            ClusterMembershipEvent.Type.MEMBER_REMOVED, event.member(), event.time());
    assertThat(receivedEvent)
        .as("the received event is the same as the one sent")
        .hasValue(expectedEvent);
  }
}
